-- VECTRIC LUA SCRIPT

--require("mobdebug").start()
require "strict"


--[[Update 1.1: SeanM     27/10/2015 Version 1.1 ready for mass public release]]
--[[Update 1.2: SeanM     02/11/2015 added slide up on instructions for users with small screens]]
--[[Update 1.2: SeanM     15/13/2016 added Error messages for invalid values]]
--[[Update 1.4: MohamedM  21/08/2019 Use ToolDBId instead of tool name]]

g_version = "1.4"
g_html_file = "Texture_Toolbox.html"

--Set options
  options = {}
  options.ToolpathName = "Texture Toolbox"
  options.ChangeRate = 5
  options.ChangeRateMax = 5
  options.WarpGradualX = 10
  options.Rotate = 1
  options.RotateDegrees = 180.0
  options.RotateEvery = 2
  options.WarpGradualY = 2
  options.WarpRandomX = 10
  options.WarpRandomY = 2
  options.MinDepth = 0
  options.MaxDepth = 1
  options.contourcount = 0
  options.AlternateNum = 2
  options.windowWidth = 734
  options.windowHeight = 876
  options.Path = ""

--[[--------------------Main Function------------------------------------
|
|     Checks if job exists, checks if objects have been selected on job, sets dialog.
|
|
--]]

function main(path)
--Check Version of software is V8
local version = GetAppVersion()
  if version < 8 then
    MessageBox("Your software needs an upgrade to V8.0 before using this gadget")
    return false
  end  
  local job = VectricJob() 
 if not job.Exists then
      MessageBox("No job created, please start a new job and run the gadget")
      return false
 end
 local selection = job.Selection
 if selection.IsEmpty then
       MessageBox("No objects selected, please select the outer most vector and re-run the gadget")
       return false
 end
 
 local pos = selection:GetHeadPosition()
 local isgroup
 isgroup,pos = selection:GetNext(pos)
 if isgroup.ClassName ~= "vcCadObjectGroup" then
   MessageBox("This gadget was designed to be used with Vectors created using the Vector Texture Tool, if you wish to use your own vectors, you must group them first, then run the gadget")
   return false
 end
  --load options from registry
LoadDefaults(options)
 
 -- Display the dialog
    local displayed_dialog = -1
    while displayed_dialog == -1 do
      displayed_dialog = DisplayDialog(path,options)
    end
    
    if not displayed_dialog then 
      return false
    end
--Gather all contours and spans and put them into an array table VSTable  
local VSTable = GetContoursAndSpans(selection)
--Create the 3D vectors 
local ToolpathVectors, contourtable = CreateToolpathContours(VSTable,job)
--Create external toolpath using created vectors.
local success = CreateExternalToolpath(ToolpathVectors)
  
job:Refresh2DView()
 
return true
end

--This function creates the toolpath from the given vectors we have created
function CreateExternalToolpath(ToolpathVectors)
  local pos_data = ToolpathPosData()
  local toolpath_options = ExternalToolpathOptions()
  toolpath_options.StartDepth = 0.0
  toolpath_options.CreatePreview = true
  local toolpath = ExternalToolpath(options.ToolpathName,options.Tool,pos_data,toolpath_options,ToolpathVectors)
  	if toolpath:Error() then
      DisplayMessageBox("Error creating toolpath")
      return false
    end

	local toolpath_manager = ToolpathManager()
	local success = toolpath_manager:AddExternalToolpath(toolpath)
	return success
end


function CreateToolpathContours(VSTable,job)
  --create contour table for debugging point data
  local contourtable = {}
  --create contourgroup to add finished contours too for the toolpath
  local ContourGroup = ContourGroup(true)
  --max amount of spans divided by the change rate specified, for equal length changes.
  --local roundspansper = math.floor(options.contourcount / options.ChangeRate)
    --for loop to cycle through the contours table
  for a =1,#VSTable do
    --create new contour
    local Contour = Contour(0.0)
    --get amount of spans for current contour
    local spancount = #VSTable[a][2]
    --set current depth for cut as the minimum depth for first initial cut
    local currentdepth = options.MinDepth
    -- calculate the Changerate for this particular vector
    local ChangeRate= math.random(options.ChangeRate,options.ChangeRateMax)
    --max amount of spans divided by the change rate specified, for equal length changes.
    local roundspansper = math.floor(options.contourcount / ChangeRate)
    -- for loop which will cycle through the change rate and equally calculate the depth per span needed
    for b = 1, ChangeRate do
      --calculate the depth per span
      local depthperspan = (-options.MaxDepth - -options.MinDepth) / roundspansper
      --set variable for spans left to process
      local spancounttodo  
      --check to see if the span count of the contour is less than the equally calculated amount of spans to process
      if spancount < roundspansper then
        --if so then set the current amount of spans to do the the amount of spans remaining
        spancounttodo = spancount
      elseif b == ChangeRate and spancount > roundspansper then
        spancounttodo = spancount
      else
        --otherwise process the equal amount calculated
        spancounttodo = roundspansper
      end
      --for loop to start processing the spancounttodo we just calculated.
      for c = 1, spancounttodo do
        --if statement to determine if the current change rate number is odd or even, 
        --we use this to determine wether we are cutting into or out of the material. odd cuts into
        if (b % 2 == 0) then
            --if the new contour we are creating has no spans, create the first point
            if Contour.Count == 0 then
               Contour:AppendPoint(VSTable[a][2][spancount].StartPoint3D.X, VSTable[a][2][spancount].StartPoint3D.Y, -currentdepth)
               spancount = spancount - 1
            else 
               --else create the points as normal
               currentdepth = currentdepth- (-depthperspan)
               Contour:LineTo(VSTable[a][2][spancount].EndPoint3D.X,VSTable[a][2][spancount].EndPoint3D.Y, -currentdepth) 
               spancount = spancount - 1      
            end
            
        else
            --if the new contour we are creating has no spans, create the first point
            if Contour.Count == 0 then
               Contour:AppendPoint(VSTable[a][2][spancount].StartPoint3D.X, VSTable[a][2][spancount].StartPoint3D.Y, -currentdepth)
            else
                --else create the points as normal
                currentdepth = currentdepth- depthperspan
                Contour:LineTo(VSTable[a][2][spancount].EndPoint3D.X,VSTable[a][2][spancount].EndPoint3D.Y,-currentdepth) 
                spancount = spancount - 1
            end
          
        end
        
      end
      
      
    end
      --After we have finished calculating the contour, add the contour to the job
      --AddContoursToJob(Contour,"Custom Texture",job)  
      ContourGroup:AddTail(Contour)   
  end
  return ContourGroup,contourtable
  
end
      
      

  
 
 --[[This function gets the contours and one by one edits them according to the users specified
     settings. then this function will also collect all of the span data for every contour 
     ready for the next function where we then modify the Z depth point for each]]
function GetContoursAndSpans(selection)
 local VSTable = {}
 local posgroup = selection:GetHeadPosition()
 local cadcontourgroup
    while (posgroup ~= nil) do 
        cadcontourgroup,posgroup = selection:GetNext(posgroup)
        local pos1 = cadcontourgroup:GetHeadPosition()
        local cadcontour
        local softwarecontourcount = cadcontourgroup.Count
        local contourcount
        local changeperlinex = (options.WarpGradualX / softwarecontourcount)
        local changecurrentx = changeperlinex
        local changeperliney = (options.WarpGradualY / softwarecontourcount)
        local changecurrenty = changeperliney
        local num = 0
        local counter = 0
        local reversecount = 0
        local altercount = 0
        local rotatecount = 0
            while (pos1 ~= nil) do 
                counter = counter+1
                num = counter
                cadcontour,pos1 = selection:GetNext(pos1)
                local contour = cadcontour:GetContour()
 ------------------------------------------------------------------------------- 
 --local function to make the current contour direction flow clockwise or anti clockwise
                 local function DirectionCW(answer,contour)
                      if answer then
                          if not contour.IsCW then
                              table.insert(VSTable[#VSTable],{false})
                              VSTable[counter][1]:Reverse()
                          else
                              table.insert(VSTable[#VSTable],{false})
                          end      
                      else
                          if not contour.IsCCW then
                              table.insert(VSTable[#VSTable],{true})
                              VSTable[counter][1]:Reverse()
                          else
                              table.insert(VSTable[#VSTable],{true})
                          end       
                      end
                  end
                  
                ------------------------------------------------------
                --local function used to rotate contours
                local function rotatecontour(VSTable,num)
                  local contour = VSTable[num][1]
                  local bbox = contour.BoundingBox2D
                  local midx = bbox.Center.X
                  local midy = bbox.Center.Y
                  local zeropoint = Point2D(0.0,0.0)
                  local transvector = Vector2D(midx,midy)
                  local translateback = TranslationMatrix2D(-transvector)
                  local translateto = TranslationMatrix2D(transvector)
                  contour:Transform(translateback)
                  local rotate = RotationMatrix2D(zeropoint, options.RotateDegrees)  
                  contour:Transform(rotate)
                  contour:Transform(translateto)
                  VSTable[num][1] = contour
                end
               ----------------------------------------------------------------------------------  
               -- local function to breakdown the contour into individual spans
               local function GetSpans()
                  local pos2 = VSTable[#VSTable][1]:GetHeadPosition()
                  local span
                    while (pos2~= nil) do 
                      span,pos2 = selection:GetNext(pos2)
                      table.insert(VSTable[#VSTable][2],span)
                    end         
               end
               -----------------------------------------------------------------------------------
               --Options to start manipulating the contours start below here, 
               --this is where to add more
               -----------------------------------------------------------------------------------
              -- check to see if contour is not empty, if it is not then perform operations
              if not contour.IsEmpty then
                --insert contour into table
                table.insert(VSTable,{contour:Clone()})
                table.insert(VSTable[#VSTable],{})
              --------------------------------------------------------  
                --check if user has specified to rotate the contour
                if options.Rotate == 2 then
                   rotatecount = rotatecount +1
                      if rotatecount >= options.RotateEvery then
                         rotatecontour(VSTable,counter)
                         rotatecount = 0
                      end
                end 
                ------------------------------------------------------      
                      
                --check the direction option chosen by the user then modify      
                if options.Type == 1 then
                   DirectionCW(true,contour)
                elseif options.Type == 2 then
                    if (num % 2 == 0) then
                       DirectionCW(true,contour)
                    else
                        DirectionCW(false,contour)
                    end
                elseif options.Type == 3 then
                    num = math.random(1,100)
                    if (num % 2 == 0) then
                       DirectionCW(true,contour)
                    else
                        DirectionCW(false,contour)
                    end
                elseif options.Type == 4 then
                    reversecount = reversecount + 1
                    altercount = altercount + 1
                    if reversecount <= options.AlternateNum then
                        if (altercount % 2 == 0) then
                           DirectionCW(true,contour)
                        else
                            DirectionCW(false,contour)
                        end
                    else
                        if VSTable[counter-1][1].IsCW then
                           DirectionCW(true,contour)
                           reversecount = 0
                           altercount = altercount + 1
                        else
                           DirectionCW(false,contour)
                           reversecount = 0
                           altercount = altercount + 1
                        end
                    end
                end
                -----------------------------------------------------      
                       --Get spans for the contour for contour building purposes
                      GetSpans()
                      contourcount = contour.Count
                      if contourcount > options.contourcount then
                        options.contourcount = contourcount
                      end
                    
              end
            end
          
    end
  return VSTable
end


--[[--------------Add contours to job/layer-----]]
function AddContoursToJob(contour,layername,job)
  local layer
  local object = CreateCadContour(contour)
      if type(layername) == "string" then
          layer = job.LayerManager:GetLayerWithName(layername)
      else
          layer = layername
      end
      if layer:AddObject(object,true) then
        return object
      end 
end



--[[  -------------- DisplayDialog --------------------------------------------------  
|
|  Display the dialog
|
]]
function DisplayDialog(path, options)
local block = MaterialBlock()
options.Path = path
local html_path = "file:" ..path .. "\\" .. g_html_file
local dialog = HTML_Dialog(false,html_path,options.windowWidth,options.windowHeight,"Texture Toolbox")
--Add preset values
dialog:AddDoubleField("ChangeRate",options.ChangeRate)
dialog:AddDoubleField("ChangeRateMax",options.ChangeRateMax)
dialog:AddDoubleField("MinDepth",options.MinDepth)
dialog:AddDoubleField("MaxDepth",options.MaxDepth)
dialog:AddLabelField("ToolNameField", "")
dialog:AddTextField("ToolpathName", options.ToolpathName)
dialog:AddToolPicker("ToolChooseButton", "ToolNameField", options.ToolId)
dialog:AddRadioGroup("type", options.Type)
dialog:AddRadioGroup("Rotate", options.Rotate)
dialog:AddIntegerField("RotateEvery",options.RotateEvery)
dialog:AddToolPickerValidToolType("ToolChooseButton", Tool.END_MILL)
dialog:AddToolPickerValidToolType("ToolChooseButton", Tool.BALL_NOSE)
dialog:AddToolPickerValidToolType("ToolChooseButton", Tool.VBIT)
dialog:AddIntegerField("AlternateNum", options.AlternateNum)
dialog:AddLabelField("FileNameLabelSave", "File Save Status:<br/>Not Saved")
dialog:AddLabelField("FileNameLabelOpen", "File Open Status:<br/>None Opened")

-- Display the dialog
      local success = dialog:ShowDialog()

      if not success then 
        return false;
      end
  --load options from dialog
options.ToolpathName = dialog:GetTextField("ToolpathName")
if options.ToolpathName == "" then
  options.ToolpathName = "Texture Toolbox"
end
options.ChangeRate = dialog:GetDoubleField("ChangeRate")
options.ChangeRateMax = dialog:GetDoubleField("ChangeRateMax")
options.Type = dialog:GetRadioIndex("type")
options.Rotate = dialog:GetRadioIndex("Rotate")
options.RotateEvery = dialog:GetIntegerField("RotateEvery")
options.MinDepth = dialog:GetDoubleField("MinDepth")
options.MaxDepth = dialog:GetDoubleField("MaxDepth")
options.Tool = dialog:GetTool("ToolChooseButton")
options.ToolId = options.Tool.ToolDBId
options.AlternateNum = dialog:GetIntegerField("AlternateNum")
options.windowWidth  = dialog.WindowWidth
options.windowHeight = dialog.WindowHeight
if options.ChangeRate < 0 or options.MinDepth < 0 then
    DisplayMessageBox("Change Rates and Cut Depths cannot be below zero, Please change the values accordingly.")
    return -1
  end
  if options.RotateEvery <= 0 or options.AlternateNum <= 0 then
    DisplayMessageBox("Rotate Every and Alternate Until Every cannot be equal or below zero, Please change the values accordingly.")
    return -1
  end
  if options.ChangeRate > options.ChangeRateMax then
    DisplayMessageBox("Minimum Change Rate cannot be higher in value than the Maximum Change Rate, Please change the values accordingly.")
    return -1
  end
   if options.MinDepth > options.MaxDepth then
    DisplayMessageBox("Minimum Cut Depth cannot be higher in value than the Maximum Cut Depth, Please change the values accordingly.")
    return -1
  end
  if options.MinDepth > block.Thickness or options.MaxDepth > block.Thickness then
    DisplayMessageBox("PLEASE NOTE: Cut depths specified go below the depth of the material, please preview toolpaths to ensure this is the expected outcome")
  end
  if options.Tool == nil then
    DisplayMessageBox("No Tool Selected.")
    return -1
  end
SaveDefaults(options)
return true
end


--[[  -------------- LoadDefaults --------------------------------------------------  
|
|  Load defaults from the registry
|
]]    
function LoadDefaults(options)
local registry = Registry("TextureToolBox")
options.ToolpathName = registry:GetString("ToolpathName",options.ToolpathName)
options.AlternateNum = registry:GetInt("AlternateNum",options.AlternateNum)
options.windowWidth = registry:GetInt("WindowWidth",  options.windowWidth)
options.windowHeight = registry:GetInt("WindowHeight", options.windowHeight)
options.ChangeRate = registry:GetInt("ChangeRate",  options.ChangeRate)
options.ChangeRateMax = registry:GetInt("ChangeRateMax",  options.ChangeRateMax)
options.MinDepth = registry:GetDouble("MinDepth",  options.MinDepth)
options.Type = registry:GetInt("Type",1)
options.Rotate = registry:GetInt("Rotate",1)
options.RotateEvery = registry:GetInt("RotateEvery",  options.RotateEvery)
options.MaxDepth = registry:GetDouble("MaxDepth",  options.MaxDepth)
options.Tool = registry:GetString("ToolChooseButton", "")
options.ToolId = ToolDBId("TextureToolBox", "")
end

--[[  -------------- SaveDefaults --------------------------------------------------  
|
| Save defaults to registry
|
]]
function SaveDefaults(options)
  local registry = Registry("TextureToolBox")
  registry:SetInt("AlternateNum",options.AlternateNum)
  registry:SetString("ToolpathName",options.ToolpathName)
  registry:SetInt("WindowWidth",options.windowWidth)
  registry:SetInt("WindowHeight",options.windowHeight)  
  registry:SetInt("Type",options.Type)
  registry:SetInt("Rotate",options.Rotate)
  registry:SetInt("ChangeRate",options.ChangeRate) 
  registry:SetInt("ChangeRateMax",options.ChangeRateMax) 
  registry:SetDouble("MinDepth",options.MinDepth)
  registry:SetInt("RotateEvery",options.RotateEvery)
  registry:SetDouble("MaxDepth",options.MaxDepth)
  options.ToolId:SaveDefaults("TextureToolBox", "")
end

function OnLuaButton_ChooseFileButtonOpen(dialog)
  local file_dialog = FileDialog()
  file_dialog:FileOpen("texture",options.Path.."\\Saved_Settings\\","Texture Settings File (*.texture) | *.texture||") 
  if file_dialog.PathName ~= "" then
  local openfile = io.open(file_dialog.PathName,"r")
  dialog:UpdateDoubleField("ChangeRate",tonumber(openfile:read("*line"))) 
  dialog:UpdateDoubleField("ChangeRateMax",tonumber(openfile:read("*line"))) 
  dialog:UpdateRadioIndex("type",tonumber(openfile:read("*line")))  
  dialog:UpdateRadioIndex("Rotate",tonumber(openfile:read("*line"))) 
  dialog:UpdateIntegerField("RotateEvery",tonumber(openfile:read("*line"))) 
  dialog:UpdateDoubleField("MinDepth",tonumber(openfile:read("*line"))) 
  dialog:UpdateDoubleField("MaxDepth",tonumber(openfile:read("*line")))
  dialog:UpdateIntegerField("AlternateNum",tonumber(openfile:read("*line"))) 
   dialog:UpdateTextField("ToolpathName",openfile:read("*line"))
  dialog:UpdateLabelField("FileNameLabelOpen","File Save Status:<br/>Opened")
  end
  
  return true
end

function OnLuaButton_ChooseFileButtonSave(dialog)
  local file_dialog = FileDialog()
  file_dialog:FileSave("texture",options.Path.."\\Saved_Settings\\","Texture Settings File (*.texture) | *.texture||")
  local filepath = file_dialog.PathName
  if file_dialog.PathName ~= "" then
  local file = io.output(filepath,"w")    
  file:write(dialog:GetDoubleField("ChangeRate"),"\n")
  file:write(dialog:GetDoubleField("ChangeRateMax"),"\n")
  file:write(dialog:GetRadioIndex("type"),"\n")
  file:write(dialog:GetRadioIndex("Rotate"),"\n")
  file:write(dialog:GetIntegerField("RotateEvery"),"\n")
  file:write(dialog:GetDoubleField("MinDepth"),"\n")
  file:write(dialog:GetDoubleField("MaxDepth"),"\n")
  file:write(dialog:GetIntegerField("AlternateNum"),"\n")
  file:write(dialog:GetTextField("ToolpathName"),"\n")  
  file:close()  
  dialog:UpdateLabelField("FileNameLabelSave","File Save Status:<br/>Saved")
  end
  return true
end